



if not exist %PREFIX%\\Library\\include\\parquet\\api\\reader.h exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\bin\\parquet.dll exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\lib\\parquet.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\lib\\parquet_static.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
